"""Use k-means for data compression."""
import os

import imageio.v2 as imageio
import matplotlib.pyplot as plt
import numpy as np
from sklearn.cluster import MiniBatchKMeans


def load_image(data_path: str) -> np.ndarray:
    """Load image from jpeg file into numpy.

    Args:
        data_path (str): Path to data file.

    Returns:
        np.ndarray: Array of image data of shape (hight, width, number of color channels).
    """
    img = imageio.imread(os.path.join(data_path, "saint_sulpice.jpg"))
    return np.asarray(img.astype(np.float32)) / 256.0


def compress_colorspace(img: np.ndarray, k: int) -> np.ndarray:
    """Compress data using k-means clustering.

    Args:
        img (np.ndarray): Numpy array of image data of shape
            (height, width, number of color channels).
        k (int): Number of clusters. Must be smaller than (height * weight) + 1 of img.

    Returns:
        np.ndarray: Compressed image array of the shape img.shape.
    """
    # 2. reshape input image into (width*height, 3) to perform clustering on colors
    img_shape = img.shape
    img = img.reshape(-1, 3)
    # 3. use `MiniBatchKMeans` to cluster image into k clusters
    batch_kmeans = MiniBatchKMeans(n_clusters=k, n_init=3).fit(img)
    pred = batch_kmeans.predict(img)
    # create new image by assigning each pixel in input image to cluster center it belongs to
    new_img = batch_kmeans.cluster_centers_[pred]
    # 4. return compressed image reshaped back into original shape
    return new_img.reshape(img_shape)


if __name__ == "__main__":
    """Perform data compression on image with k-means and plot result."""
    input_img = load_image("./data/images")

    # 1. print image dimensions
    print(" Image dimensions:", input_img.shape)

    # 5. use `compress_colorspace` to compress image
    # for each k in {2,8,64,256} and plot each result via imshow;
    # set value of k as title
    values = [2, 8, 64, 256]
    for index, k in enumerate(values):
        print(f"Index: {index}, k: {k}")

        compressed_img = compress_colorspace(input_img, k)
        plt.title(f"k= {k}")
        plt.imshow(compressed_img)
        plt.tight_layout()
        plt.axis("off")
        plt.savefig(
            "./figures/compressed_image_{}_k{}.png".format(index, k),
            bbox_inches="tight",
        )
        plt.show()
